from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Literal, TypeAlias

from .. import MultihostHost, MultihostUtility
from ..ssh import SSHLog

__all__ = ["LinuxFirewalld"]


ProtocolSpec: TypeAlias = Literal["tcp", "udp"]
"""Firewall protocol specification."""


PortSpec: TypeAlias = int | tuple[int, ProtocolSpec]
"""Firewall port specification."""


ServiceSpec: TypeAlias = str | PortSpec
"""Firewall service specification: service name | port | (port, "tcp" | "udp")."""


class GenericFirewall(ABC, MultihostUtility):
    """
    Configure host firewall.

    All changes are automatically reverted when a test is finished.
    """

    @abstractmethod
    def accept(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Accept incoming traffic on given port.

        .. code-block:: python

            firewall.accept(389)  # missing protocol defaults to "tcp"
            firewall.accept((389, "tcp"))
            firewall.accept([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    @abstractmethod
    def drop(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Drop incoming traffic on given port.

        .. code-block:: python

            firewall.drop(389)  # missing protocol defaults to "tcp"
            firewall.drop((389, "tcp"))
            firewall.drop([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        pass

    def parse_port_spec(self, spec: PortSpec) -> tuple[int, ProtocolSpec]:
        """
        Parse port specification into (port, protocol) tuple.

        :raises TypeError: When invalid type is given.
        :return: (port, protocol) tuple)
        :rtype: tuple[int, Literal["tcp", "udp"]]
        """
        port: int = 0
        protocol: ProtocolSpec = "tcp"

        if isinstance(spec, int):
            port = spec
            protocol = "tcp"
        elif isinstance(spec, tuple):
            if list(map(type, spec)) == [int, str]:
                port = spec[0]
                protocol = spec[1]
            else:
                raise TypeError(f"Unexpected type: tuple{list(map(type, spec))}")
        else:
            raise TypeError(f"Unexpected type: {type(spec)}")

        if protocol not in ["tcp", "udp"]:
            raise ValueError(f"Unexpected protocol: {protocol}")

        return (port, protocol)


class LinuxFirewalld(GenericFirewall):
    """
    Configure firewall using firewalld.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)

        self._priority: int = 30000
        """Next priority for accept/reject/drop rules."""

    def teardown_when_used(self):
        """
        Revert all firewall changes.

        :meta private:
        """
        self.host.ssh.exec(["firewall-cmd", "--reload"])
        super().teardown_when_used()

    def accept(self, port_or_service: ServiceSpec | list[ServiceSpec]) -> None:  # type: ignore[override]
        """
        Accept incoming traffic on given port or service.

        .. code-block:: python

            firewall.accept(389)  # missing protocol defaults to "tcp"
            firewall.accept((389, "tcp"))
            firewall.accept("ldap")
            firewall.accept(["ldap", "ldaps"])

        :param port_or_service: Port (integer), (port, protocol) tuple or service (string).
        :type port_or_service: ServiceSpec | list[ServiceSpec]
        """
        self.__add_action(port_or_service, action="accept")

    def reject(self, port_or_service: ServiceSpec | list[ServiceSpec]) -> None:
        """
        Reject incoming traffic on given port or service.

        .. code-block:: python

            firewall.reject(389)  # missing protocol defaults to "tcp"
            firewall.reject((389, "tcp"))
            firewall.reject("ldap")
            firewall.reject(["ldap", "ldaps"])

        :param port_or_service: Port (integer), (port, protocol) tuple or service (string).
        :type port_or_service: ServiceSpec | list[ServiceSpec]
        """
        self.__add_action(port_or_service, action="reject")

    def drop(self, port_or_service: ServiceSpec | list[ServiceSpec]) -> None:  # type: ignore[override]
        """
        Drop incoming traffic on given port or service.

        .. code-block:: python

            firewall.drop(389)  # missing protocol defaults to "tcp"
            firewall.drop((389, "tcp"))
            firewall.drop("ldap")
            firewall.drop(["ldap", "ldaps"])


        :param port_or_service: Port (integer), (port, protocol) tuple or service (string).
        :type port_or_service: ServiceSpec | list[ServiceSpec]
        """
        self.__add_action(port_or_service, action="drop")

    def add_rich_rule(self, rule: str) -> None:
        """
        Add rich rule.

        :param rule: Firewalld rich rule.
        :type rule: str
        """
        self.logger.info(f'Firewalld: adding rich rule "{rule}" on {self.host.hostname}')
        self.host.ssh.exec(["firewall-cmd", "--add-rich-rule", rule], log_level=SSHLog.Error)

    def remove_rich_rule(self, rule: str) -> None:
        """
        Remove rich rule.

        :param rule: Firewalld rich rule.
        :type rule: str
        """
        self.logger.info(f'Firewalld: removing rich rule  "{rule}" on {self.host.hostname}')
        self.host.ssh.exec(["firewall-cmd", "--remove-rich-rule", rule], log_level=SSHLog.Error)

    def __add_action(
        self, port_or_service: ServiceSpec | list[ServiceSpec], *, action: Literal["accept", "reject", "drop"]
    ) -> None:
        def __add_port(port: int, protocol: str):
            self.add_rich_rule(f'rule priority="{self._priority}" port port="{port}" protocol="{protocol}" {action}')

        def __add_service(name: str):
            self.add_rich_rule(f'rule priority="{self._priority}" service name="{name}" {action}')

        if not isinstance(port_or_service, list):
            port_or_service = [port_or_service]

        for item in port_or_service:
            if isinstance(item, str):
                __add_service(item)
                continue

            port, protocol = self.parse_port_spec(item)
            __add_port(port, protocol)

        self._priority -= 1


class WindowsFirewall(GenericFirewall):
    """
    Configure Windows firewall.

    All changes are automatically reverted when a test is finished.
    """

    def __init__(self, host: MultihostHost) -> None:
        """
        :param host: Remote host instance.
        :type host: MultihostHost
        """
        super().__init__(host)
        self._changed: bool = False
        """Did we change anything?"""

        self._rules: list[str] = []
        self._backup: str = "C:\\.mh_firewall.bak.wfw"

    def setup_when_used(self):
        """
        Create a backup of current firewall configuration.

        :meta private:
        """
        super().setup_when_used()
        self.logger.info(f"Windows Firewall: creating backup at '{self._backup}' on {self.host.hostname}")
        self.host.ssh.run(
            f"Remove-Item {self._backup}; netsh advfirewall export {self._backup}", log_level=SSHLog.Error
        )

    def teardown_when_used(self):
        """
        Revert all firewall changes.

        :meta private:
        """
        self.logger.info(f"Windows Firewall: restoring from '{self._backup}' on {self.host.hostname}")
        self.host.ssh.run(f"netsh advfirewall reset; netsh advfirewall import {self._backup}", log_level=SSHLog.Error)
        super().teardown_when_used()

    def accept(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Accept incoming traffic on given port.

        .. code-block:: python

            firewall.accept(389)  # missing protocol defaults to "tcp"
            firewall.accept((389, "tcp"))
            firewall.accept([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        self._add_port_rule(port, action="allow")

    def drop(self, port: PortSpec | list[PortSpec]) -> None:
        """
        Drop incoming traffic on given port.

        .. code-block:: python

            firewall.drop(389)  # missing protocol defaults to "tcp"
            firewall.drop((389, "tcp"))
            firewall.drop([389, 636])

        :param port: Port (integer), (port, protocol) tuple.
        :type port: PortSpec | list[PortSpec]
        """
        self._add_port_rule(port, action="block")

    def _add_port_rule(self, port: PortSpec | list[PortSpec], *, action: Literal["allow", "block"]) -> None:
        if not isinstance(port, list):
            port = [port]

        for item in port:
            port, protocol = self.parse_port_spec(item)

            name = f"mh/{action}/{port}/{protocol}"
            match action:
                case "allow":
                    opposite = f"mh/block/{port}/{protocol}"
                case "block":
                    opposite = f"mh/allow/{port}/{protocol}"
                case _:
                    raise ValueError(f"Unknown action: {action}")

            remove = f"Remove-NetFirewallRule -DisplayName '{opposite}'; " if opposite in self._rules else ""
            add = f"New-NetFirewallRule -DisplayName '{name}' -Action {action} -Protocol {protocol} -LocalPort {port}"

            self.logger.info(f'Windows Firewall: {action} "{port}/{protocol}" on {self.host.hostname}')
            self.host.ssh.run(f"{remove}{add}", log_level=SSHLog.Error)
            self._rules.append(name)



from __future__ import annotations

from types import SimpleNamespace
from typing import Generator

import pytest

from .data import MultihostItemData
from .logging import MultihostLogger
from .multihost import MultihostConfig, MultihostDomain, MultihostHost, MultihostRole
from .topology import Topology, TopologyDomain


class MultihostFixture(object):
    """
    Multihost object provides access to underlaying multihost configuration,
    individual domains and hosts. This object should be used only in tests
    as the :func:`mh` pytest fixture.

    Domains are accessible as dynamically created properties of this object,
    hosts are accessible by roles as dynamically created properties of each
    domain. Each host object is instance of specific role class based on
    :mod:`~pytest_mh.MultihostRole`.

    .. code-block:: yaml
        :caption: Example multihost configuration

        domains:
        - id: test
          hosts:
          - name: client
            hostname: client.test
            role: client

          - name: ldap
            hostname: master.ldap.test
            role: ldap

    The configuration above creates one domain of id ``test`` with two hosts.
    The following example shows how to access the hosts:

    .. code-block:: python
        :caption: Example of the MultihostFixture object

        def test_example(mh: MultihostFixture):
            mh.test            # -> namespace containing roles as properties
            mh.test.client     # -> list of hosts providing given role
            mh.test.client[0]  # -> host object, instance of specific role
    """

    def __init__(
        self, request: pytest.FixtureRequest, data: MultihostItemData, multihost: MultihostConfig, topology: Topology
    ) -> None:
        """
        :param request: Pytest request.
        :type request: pytest.FixtureRequest
        :param data: Multihost item data.
        :type data: MultihostItemData
        :param multihost: Multihost configuration.
        :type multihost: MultihostConfig
        :param topology: Multihost topology for this request.
        :type topology: Topology
        """

        self.data: MultihostItemData = data
        """
        Multihost item data.
        """

        self.request: pytest.FixtureRequest = request
        """
        Pytest request.
        """

        self.multihost: MultihostConfig = multihost
        """
        Multihost configuration.
        """

        self.logger: MultihostLogger = multihost.logger
        """
        Multihost logger.
        """

        self._paths: dict[str, list[MultihostRole] | MultihostRole] = {}

        for domain in self.multihost.domains:
            if domain.id in topology:
                setattr(self, domain.id, self._domain_to_namespace(domain, topology.get(domain.id)))

    def _domain_to_namespace(self, domain: MultihostDomain, topology_domain: TopologyDomain) -> SimpleNamespace:
        ns = SimpleNamespace()
        for role_name in domain.roles:
            if role_name not in topology_domain:
                continue

            count = topology_domain.get(role_name)
            roles = [domain.create_role(self, host) for host in domain.hosts_by_role(role_name)[:count]]

            self._paths[f"{domain.id}.{role_name}"] = roles
            for index, role in enumerate(roles):
                self._paths[f"{domain.id}.{role_name}[{index}]"] = role

            setattr(ns, role_name, roles)

        return ns

    def _lookup(self, path: str) -> MultihostRole | list[MultihostRole]:
        """
        Lookup host by path. The path format is ``$domain.$role``
        or ``$domain.$role[$index]``

        :param path: Host path.
        :type path: str
        :raises LookupError: If host is not found.
        :return: The role object if index was given, list of role objects otherwise.
        :rtype: MultihostRole | list[MultihostRole]
        """

        if path not in self._paths:
            raise LookupError(f'Name "{path}" does not exist')

        return self._paths[path]

    @property
    def _hosts_and_roles(self) -> list[MultihostHost | MultihostRole]:
        """
        :return: List containing all hosts and roles available for current test case.
        :rtype: list[MultihostHost | MultihostRole]
        """
        roles: list[MultihostRole] = [x for x in self._paths.values() if isinstance(x, MultihostRole)]
        hosts: list[MultihostHost] = [x.host for x in roles]

        return list(set([*hosts, *roles]))

    def _setup(self) -> None:
        """
        Setup multihost. A setup method is called on each host and role
        to initialize the test environment to expected state.
        """
        setup_ok: list[MultihostHost | MultihostRole] = []
        for item in self._hosts_and_roles:
            try:
                item.setup()
            except Exception:
                # Teardown hosts and roles that were successfully setup before this error
                for i in reversed(setup_ok):
                    i.teardown()
                raise

            setup_ok.append(item)

    def _teardown(self) -> None:
        """
        Teardown multihost. The purpose of this method is to revert any changes
        that were made during a test run. It is automatically called when the
        test is finished.
        """
        errors = []
        for item in reversed(self._hosts_and_roles):
            try:
                # Try to collect artifacts from host before the role is teared down.
                # We need to do it before the role object is teardown as it may
                # potentially remove some of the requested artifacts.
                if isinstance(item, MultihostRole):
                    self._collect_artifacts(item.host)

                item.teardown()
            except Exception as e:
                errors.append(e)

        if errors:
            raise Exception(errors)

    def _collect_artifacts(self, host: MultihostHost) -> None:
        """
        Collect test artifacts that were requested by the multihost configuration.

        :param host: Host object where the artifacts will be collected.
        :type host: MultihostHost
        """
        dir = self.request.config.getoption("mh_artifacts_dir")
        mode = self.request.config.getoption("mh_collect_artifacts")
        if mode == "never" or (mode == "on-failure" and self.data.outcome != "failed"):
            return

        host.collect_artifacts(f"{dir}/{self.request.node.name}")

    def __enter__(self) -> MultihostFixture:
        self._setup()
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> None:
        self._teardown()


@pytest.fixture(scope="function")
def mh(request: pytest.FixtureRequest) -> Generator[MultihostFixture, None, None]:
    """
    Pytest multihost fixture. Returns instance of :class:`MultihostFixture`.
    When a pytest test is finished, this fixture takes care of tearing down the
    :class:`MultihostFixture` object automatically in order to clean up after
    the test run.

    .. note::

        It is preferred that the test case does not use this fixture directly
        but rather access the hosts through dynamically created role fixtures
        that are defined in ``@pytest.mark.topology``.

    :param request: Pytest's ``request`` fixture.
    :type request: pytest.FixtureRequest
    :raises ValueError: If not multihost configuration was given.
    :yield: MultihostFixture
    """

    data: MultihostItemData | None = MultihostItemData.GetData(request.node)
    if data is None:
        nodeid = f"{request.node.parent.nodeid}::{request.node.originalname}"
        raise ValueError(f"{nodeid}: mh fixture requested but no multihost configuration was provided")

    if data.multihost is None:
        raise ValueError("data.multihost must not be None")

    if data.topology_mark is None:
        raise ValueError("data.topology_mark must not be None")

    with MultihostFixture(request, data, data.multihost, data.topology_mark.topology) as mh:
        yield mh
